#include <stdlib.h>
#include <ctype.h>

#include "lpcprog.h"
#include "srec.h"

uint32_t ctoi(char c)
{
    c = toupper(c);

    if (c>='0' && c<='9')
        return c-'0';
    if (c>='A' && c<='F')
        return 10 + c - 'A';

    printf("invalid char %c\n", c);
    return 0;
}

uint32_t stoi(char *c, int len)
{
    uint32_t v = 0;

    for (int i = 0; i < len; i++)
        v = (v<<4) + ctoi(c[i]);

    return v;
}

void decodehex(char *in, int nbytes, char *data)
{
    for (int i = 0; i < nbytes; i++)
    {
        data[i] = stoi(&in[i*2], 2);
    }
}

objdata_t *srec_parse(const char *path, int chunksize)
{
    FILE *f = fopen(path, "r");
    if (f == NULL)
        return NULL;

    objdata_t *obj = (objdata_t*) calloc(1, sizeof(objdata_t));
    
    char line[1024];
    chunk_t *current = NULL;
    char data[100];

    // http://www.amelek.gda.pl/avr/uisp/srecord.htm
    while (fgets(line, 1000, f)!=NULL)
    {
        if (line[0] != 'S')
            continue;
        int addr;
        int count = stoi(&line[2], 2);
       
        // check the checksum
        int chk = 0;
        for (int i = 0; i < count+1; i++) {
            int bval = stoi(&line[2+i*2],2);
            chk = (chk + bval) & 0xff;
        }

        if (chk != 0xff)
            printf("BAD SRECORD CHECKSUM!  %s\n", line);

        switch (line[1])
        {
        case '0':
            // module name, version, revision, text comment
            continue;

        case '1':
            addr = stoi(&line[4], 4);
            count -= 3;
            decodehex(&line[8], count, data);
            break;

        case '2':
            addr = stoi(&line[4], 6);
            count -= 4;
            decodehex(&line[10], count, data);
            break;

        case '3':
            addr = stoi(&line[4], 8);
            count -= 5;
            decodehex(&line[12], count, data);
            break;

        case '5': 
            // count of S1,S2,S3 records previously transmitted
            continue;

        case '7':
            obj->startaddr = stoi(&line[4], 8);
            continue;
        case '8':
            obj->startaddr = stoi(&line[4], 6);
            continue;
        case '9':
            obj->startaddr = stoi(&line[4], 4);
            continue;

        default:
        case '4':
        case '6':
            printf("undefined S record! S%c\n", line[1]);
            continue;
        }

//        printf("%08x %d : ", addr, count);

        for (int i = 0; i < count; i++)
        {
            int thisaddr = addr + i;

            if (current == NULL || thisaddr >= (current->addr + current->datalen))
            {
                chunk_t *newchunk = chunk_create(chunksize);
                obj->nchunks ++;

                if (current != NULL)
                    current->next = newchunk;

                current = newchunk;
                current->addr = thisaddr - (thisaddr & (chunksize - 1));

                if (obj->chunks == NULL)
                    obj->chunks = current;

//                printf("CHUNK %08X %d\n", current->addr, current->datalen);
            }

            current->data[thisaddr - current->addr] = data[i];
        }
    }
    
    return obj;
}
